/*
 * graphics.h
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (c) 2006 Andrzej Zaborowski  <balrog@zabor.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#ifndef __graphics_h_
# define __graphics_h_

# define RES_X		0x140
# define RES_Y		0x140
# define FRAME_OFF	0x200
# define RECT_LEFT	0x020
# define RECT_RIGHT	0x120
# define RECT_TOP	0x118
# define RECT_BOTTOM	0x120

# define BI_PP		8
# define BY_PP		(BI_PP / 8)
# define FB_BASE	(void *) fb_base
# define FB_FRAME	(FB_BASE + FRAME_OFF)

# if BI_PP == 8
#  define SET_PIX(x, y, c)	\
	*(unsigned char *) (FB_FRAME + (y) * RES_X + (x)) = c
# endif
# if BI_PP == 16
#  define SET_PIX(x, y, c)	\
	*(unsigned short *) (FB_FRAME + ((y) * RES_X + (x)) * 2) = c
# endif
# define LINE_HORZ(x, y, z, c)	\
	for (v0 = x; v0 <= z; v0 ++)	\
		SET_PIX(v0, y, c)
# define LINE_VERT(x, y, z, c)	\
	for (v0 = y; v0 <= z; v0 ++)	\
		SET_PIX(x, v0, c)
# define FRAME(a, b, c, d, e)	\
	LINE_HORZ(a, b, c, e);	\
	LINE_HORZ(a, d, c, e);	\
	LINE_VERT(a, b, d, e);	\
	LINE_VERT(c, b, d, e)
# define COUNTER_INIT		\
	*(unsigned short *) (FB_BASE + 2) = 0x0f81;	\
	*(unsigned short *) (FB_BASE + 4) = 0x0000;	\
	FRAME(RECT_LEFT, RECT_TOP, RECT_RIGHT, RECT_BOTTOM, 2)
# define COUNTER_UPD(p, s, max)	\
	v1 = RECT_LEFT + 2 + (((p) - (s)) >> 8) *	\
		(RECT_RIGHT - RECT_LEFT - 4) / (((max) - (s)) >> 8);	\
	LINE_VERT(v1, RECT_TOP + 2, RECT_BOTTOM - 2, 1)

#endif	/* __graphics_h_ */
