/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <MemGlue.h>
#include <MemoryMgr.h>
#include <string.h>
#include <stdio.h>

#include "garux.h"


UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
  int i = 0;
  //  MemHandle defaultBootParm;
  char *kernelParameters = /*"root=/dev/mmcblk0p1 mem=32m console=tty0 fbcon=map:0"*/"Roger red-hat had a red ball.  A little white dog passed by.";
  EventType event;
  UInt16 error_code;
  char deviceID[4];
  MemPtr kernel;
  MemHandle asmloaderHandle;
  void *asmloader;
  MemHandle pageHandle;
  void *page;
  FieldType *kernelParametersField;
  MemHandle splashRes;
  BitmapType *splashBmp;
  Err retcode;
  UInt32 creator=3007;
  UInt16 featureNum=35;
  UInt16 irport = 0;
 
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    // Let's get the device ID in order to check wether we're on a T|E
    FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, (UInt32 *)&deviceID);

    if (1) {
      FrmGotoForm(MainForm);
      while (1) {
	EvtGetEvent(&event, 100);
	if (SysHandleEvent(&event))
	  continue;
	if (MenuHandleEvent(0, &event, &error_code))
	  continue;
	if (event.eType == ctlSelectEvent) {
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FldSetTextHandle(kernelParametersField, 0);
	    FrmCloseAllForms();
	    return 0;
	  }
	  else if (event.data.ctlSelect.controlID == LinuxButton) {
	    /* User chose to load Linux */

	    // Reserve a chunk of memory for the kernel
	    // We allocate 1Mb.
/*	retcode = FtrPtrNew (creator,featureNum,0x300000,(void**)&kernel);
  if (retcode != 0) {
    printf ("Cannot allocate memory !!\n");
    SysTaskDelay( 1*SysTicksPerSecond() );
    return -1;
  }*/
	    kernel = MemGluePtrNew(0x00008000);

	    /* Fetch kernel command line */
	    strcpy(kernel, FldGetTextPtr(kernelParametersField));
	    FldSetTextHandle(kernelParametersField, 0);

	    /* Let's close the form now that we don't need the text field anymore */
	    FrmCloseAllForms();

	    /* Let's resume memory allocation */

	    asmloaderHandle = DmGetResource('armc', 0);
	    asmloader = MemHandleLock(asmloaderHandle);

	    printf("Loading kernel... at 0x%lx\n",(unsigned long)kernel);

	    if (SrmOpen(0x6972636d, 9600, &irport) != errNone)
	      printf("Cannot open port!\n");
	    if (SrmControl(irport, srmCtlIrDAEnable, 0, 0) != errNone)
	      printf("Cannot setup Ir!\n");
	    if (SrmControl(irport, srmCtlRxEnable, 0, 0) != errNone)
	      printf("Cannot setup Rx!\n");
/*	    SrmReceiveFlush(irport, 1);
	    SrmSend(irport, "Garux: Port open under PalmOS.\n", 31, &retcode);
	    SrmSendFlush(irport);
*/
	    printf("Joining kernel chunks deferred.\n");
#if 0
	    for (i = 0; i < NUM_KERNEL_PAGES; i++) {
	      pageHandle = DmGetResource('page', i);
	      page = MemHandleLock(pageHandle);
	      memcpy(kernelPointer, page, PAGE_SIZE);
	      kernelPointer += PAGE_SIZE;
	      MemHandleUnlock(pageHandle);
	      DmReleaseResource(pageHandle);
	      putchar('.');
	    }
	    printf("OK\n");
#endif

	    printf("Running assembly loader... Please wait.\n");

	    splashRes = DmGetResource(bitmapRsc, Splash);
	    splashBmp = MemHandleLock(splashRes);
	    WinPaintBitmap(splashBmp, 0, 0);
	    MemHandleUnlock(splashRes);

	    PceNativeCall(asmloader, kernel);

	    /* The following code shall not be executed */
	    MemPtrFree(kernel);
	    MemHandleUnlock (asmloaderHandle);
	    return 0;
	  }
	}
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());

	  // Initial warning
	  if (FrmAlert(StartupAlert) == 1) {
	    FrmCloseAllForms();
	    return 0;
	  }

	  // Initializing boot parameters
	  kernelParametersField = (FieldType *)FrmGetObjectPtr(FrmGetFormPtr(MainForm), FrmGetObjectIndex(FrmGetFormPtr(MainForm), KernelParam));

	  FldSetTextHandle(kernelParametersField, DmGetResource(strRsc, kernelDefaultParam));
	  FldRecalculateField(kernelParametersField, true);
          continue;
        }

	if (event.eType == appStopEvent) {
	  FldSetTextHandle(kernelParametersField, 0);
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	  FrmHandleEvent(FrmGetActiveForm(), &event);
      }
    }
    else {
      // Garux is not being run on a Tungsten E. We'll warn, and exit.
      FrmGotoForm(AltForm);
      
      while (1) {
	EvtGetEvent(&event, 100);
	SysHandleEvent(&event);
	if (event.eType == ctlSelectEvent)
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FrmCloseAllForms();
	    return 0;
	  }
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());
	  continue;
	}
	if (event.eType == appStopEvent) {
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	FrmHandleEvent(FrmGetActiveForm(), &event); 
      }
    }
    return 0;
  }
}
