/*
 * asmloader.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * Support for images above 1 MB and LCD by Andrzej Zaborowski
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PceNativeCall.h>
#include <Standalone.h>

#include "garux.h"
#include "graphics.h"

#define SET_TAG_WORD(off, value) *((unsigned long *)(0x10000100+((off)<<2)))=value
#define SET_TAG_WORD_BYTE(off, byte, value) *((char *)(0x10000100+((off)<<2)+(byte)))=value

STANDALONE_CODE_RESOURCE_ID (0);

#if 0
#define DEBUG(a) { \
    unsigned long register __temp__; \
    for(__temp__=0 ; __temp__<1024 ; __temp__+=4) { \
        *((unsigned long *)((a)+__temp__))=0xffffffff; \
    } \
}
#endif
#define DEBUG(a)

#define FB_ADDR	(omap_readw(0xfffedb02) | (omap_readw(0xfffedb04) << 16))
#define fb_base	(*(uint32_t *) 0x10000000)
#define fb_end	(*(uint32_t *) 0x10000004)

void startKernel();

unsigned long bootLinux(const void *emulStateP, char *userData68KP, Call68KFuncType *call68KFuncP) {
    unsigned long register i;
    unsigned long register j;
    unsigned long register k;
    int byte = 0;
    char c, *cmdline;

    cmdline = (char*)userData68KP;

    // Disable FIQ, IRQ
    asm volatile ("mrs r0, cpsr");
    asm volatile ("orr r0, r0, #0xc0");
    asm volatile ("msr cpsr, r0");

    // Set Page table entry for 0x10000000 to identity.
    *((unsigned long *)0x4400)=0x10000c1a;
    // Set Page table entry for 0x20000000 to identity.
    *((unsigned long *)0x4C00)=0x20000c1a;

    // Let's flush Instruction and Data cache
    asm volatile ("mov r0, #0");
    asm volatile ("mcr p15, 0, r0, c7, c7, 0");       // flush v3/v4 cache
    asm volatile ("mcr p15, 0, r0, c8, c7, 0");       // flush v4 TLB

    // Wait a bit.
    for(i=0 ; i<100000 ; i++);

    DEBUG(0x11f08000);

    // Set the ATags

    i=0;
    // ATAG_CORE
    SET_TAG_WORD(i++,0x00000002);
    SET_TAG_WORD(i++,0x54410001);

    // ATAG_MEM
    SET_TAG_WORD(i++,0x00000004);
    SET_TAG_WORD(i++,0x54410002);
    SET_TAG_WORD(i++,0x02000000);
    SET_TAG_WORD(i++,0x10000000);

    // ATAG_FB
    SET_TAG_WORD(i ++, 0x00000009);
    SET_TAG_WORD(i ++, 0xfbfbfbfb);
    SET_TAG_WORD(i ++, FB_ADDR);
    SET_TAG_WORD(i ++, 0x00000028);
    SET_TAG_WORD(i ++, 0x00000028);
    SET_TAG_WORD(i ++, 0x00000000);
    SET_TAG_WORD(i ++, 0x00000000);
    SET_TAG_WORD(i ++, 0x00000008);
    SET_TAG_WORD(i ++, 0x00000010);

    // ATAG_CMDLINE
    for (byte=0; cmdline[byte]; byte++);
    byte = (byte+3) & ~3;
    SET_TAG_WORD(i++,0x00000002+byte/4);
    SET_TAG_WORD(i++,0x54410009);
    byte = 0;
    do {
        c = *cmdline++;
        SET_TAG_WORD_BYTE(i,byte,c);
        if (++byte==4) { byte=0; i++; }
    } while (c);
    while (byte>0) {
        SET_TAG_WORD_BYTE(i,byte,0);
        if (++byte==4) { byte=0; i++; }
    }

    // ATAG_OMAP
    SET_TAG_WORD(i++,0x00000005);
    SET_TAG_WORD(i++,0x414f4d50);
    SET_TAG_WORD(i++,0x4f030005);
    SET_TAG_WORD(i++,0x030001c2);
    SET_TAG_WORD(i++,0x00000000);
   
    // ATAG_NONE
    SET_TAG_WORD(i++,0x00000000);
    SET_TAG_WORD(i++,0x00000000);

    // FIXME: We arbitrarily add 0x60 to pc and copy 2048 bytes starting
    // from there. So if any code is modified between <CRITICAL>
    // and </CRITICAL>, the offset needs to change.
    // <CRITICAL>
    asm volatile ("add %0, pc, #0x60" : "=r" (j));
    for(i=0 ; i<2048 ; i++) {
        *((char *)(0x10001000+i))=*((char *)(i+j));
    }

    DEBUG(0x11f09000);

    asm volatile ("mov lr,     #0x10000000;"
		  "add pc, lr, #0x00001000;");

    // </CRITICAL>

    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");

    // Let's flush Instruction and Data cache
    asm volatile ("mov r0, #0");
    asm volatile ("mcr p15, 0, r0, c7, c7, 0");       // flush v3/v4 cache
    asm volatile ("mcr p15, 0, r0, c8, c7, 0");       // flush v4 TLB

    // Wait a bit.
    for(i=0 ; i<100000 ; i++);
    
    // Let's disable the MMU and disable data cache.
    asm volatile ("mrc p15, 0, r0, c1, c0, 0");
    asm volatile ("bic r0, r0, #0x00002300");         // clear bits 13, 9:8 (--V- --RS)
    asm volatile ("bic r0, r0, #0x00000087");         // clear bits 7, 2:0 (B--- -CAM)
    asm volatile ("orr r0, r0, #0x00001000");         // set bit 12 (I) I-Cache
    asm volatile ("mcr p15, 0, r0, c1, c0, 0");
    
    // Wait a bit.
    for(i=0 ; i<100000 ; i++);
    
    DEBUG(0x11f0a000);

    fb_base = FB_ADDR;
    fb_end = FB_ADDR + 0x000191fe;

    /* Set up OMAP LCD controller */
    omap_writel(0x00000000, 0xfffec010);
    omap_writel(0xfe810c80, 0xfffec000);
    for (i = 0; i < 1000000; i ++);	/* Let the DMA finish last frame */
    omap_writew(0x0000, 0xfffedb00);
    omap_writew(fb_base >> 16, 0xfffedb04);
    omap_writew(fb_base & 0xffff, 0xfffedb02);
    omap_writew(fb_end >> 16, 0xfffedb08);
    omap_writew(fb_end & 0xffff, 0xfffedb06);
    omap_writel(0x00000000, 0xfffec014);
    omap_writel(0xff300000, 0xfffec00c);
    omap_writel(0x1b070d3f, 0xfffec004);
    omap_writel(0x0708013f, 0xfffec008);
    omap_writel(0x03300004, 0xfffec00c);
    omap_writel(0xfe810c81, 0xfffec000);

    /*
     * Stop generic DMA channels as they might be overwriting our
     * data in random places.  It's also a requirement for Linux.
     */

    /* First disable TC traps */
    omap_writew(0x0009, 0xfffed30c);		/* ENHANCED_TIPB_CNTL */

    /* DMA_CCR/SYNC */
    for (i = 0; i < 9; i ++)
        omap_clrw(0xfffed802 + 0x40 * i, 0, 5);
    for (i = 0; i < 1000000; i ++);

    /* Now disable the running channels.  */
    for (i = 0; i < 9; i ++)
        omap_clrw(0xfffed802 + 0x40 * i, 7, 1);	/* DMA_CCR/EN */

    /*
     * Try to switch off all unused OMAP peripherals as they will
     * just sit consuming power if they were in use by PalmOS and
     * Linux is not using them.
     */

    /* PWT */
    omap_writel(0, 0xfffb6004);
    omap_writel(0, 0xfffb6008);
    /* First LPG */
    omap_writeb(0, 0xfffbd000);
    omap_writeb(0, 0xfffbd004);
    /* Second LPG */
    omap_writeb(0, 0xfffbd800);
    omap_writeb(0, 0xfffbd804);
    /* USB Function */
    omap_writel(0, 0xfffb4018);
    /* USB Host */
    omap_writew(0, 0xfffec208);
    /* HDQ and 1-Wire Protocols */
    omap_writel(0, 0xfffbc008);
    /* Frame Adjustment Counter */
    omap_writew(0, 0xfffba800);
    omap_writew(0, 0xfffba808);
    /* uWire */
    omap_writew(0x0c00, 0xfffb3004);
    omap_writew(0x0000, 0xfffb3004);
    omap_writew(0, 0xfffb3010);
    omap_writew(0, 0xfffb3018);
#if 0
    /* MCSI1 */
    omap_writew(0, 0xe1012800);
    omap_writew(2, 0xe1012800);
    /* MCSI2 */
    omap_writew(0, 0xe1012000);
    omap_writew(2, 0xe1012000);
#endif
    /* McBSP1 (Sound output) */
    omap_writew(0, 0xe101180a);
    omap_writew(0, 0xe1011808);
    omap_writew(0x4000, 0xe1011824);
#if 0
    /* McBSP2 */
    omap_writew(0, 0xfffb100a);
    omap_writew(0, 0xfffb1008);
    omap_writew(0x4000, 0xfffb1024);
    /* McBSP3 */
    omap_writew(0, 0xe101700a);
    omap_writew(0, 0xe1017008);
    omap_writew(0x4000, 0xe1017024);
#endif
    /* Old I2C */
    omap_writew(1, 0xfffb3810);
    /* New I2C */
    omap_writew(0, 0xfffb3824);
    /* USB Host & 32-kHz oscillator & MMC */
    omap_writel((omap_readl(0xfffe1080) & 0xf07ffdfe) | 0x08000000,
            0xfffe1080);

#define v0	j
#define v1	k
    COUNTER_INIT;

    /*
     * Grep memory for our kernel pages and join them at an arbitrary
     * location in memory assuming there's nothing important there.
     * Keep the ordering of kernel chunks.
     */
    for (i = 0x10000000; i < DEST_ADDR;) {
        if (((struct kernel_page_s *) i)->magic == 0xcafebabe) {
            i += 4;
            if (*(uint8_t *) i != '0')
                continue;
            else
                *(uint32_t *) (i - 4) = 0xdeadbabe;
            j = *(uint8_t *) i ++ - '0';
            j *= 10;
            j += *(uint8_t *) i ++ - '0';
            j *= 10;
            j += *(uint8_t *) i ++ - '0';
            j *= 10;
            j += *(uint8_t *) i ++ - '0';
            j *= PAGE_SIZE - 8;
            j += DEST_ADDR;
            for (k = 8; k < PAGE_SIZE; i += 4, j += 4, k += 4)
                *(uint32_t *) j = *(uint32_t *) i;
            COUNTER_UPD(i, 0x10000000, DEST_ADDR);
        } else
            i += 4;
        if (!(i & 0xffff)) {
            COUNTER_UPD(i, 0x10000000, DEST_ADDR);
        }
    }

    // Actually copy the kernel to 0x10008000.
    for(i=0 ; i<NUM_KERNEL_PAGES*(PAGE_SIZE-8) ; i+=4) {
        *((unsigned long *)(0x10008000+i))=*((unsigned long *)(DEST_ADDR+i));
    }

    DEBUG(0x11f0b000);

    // Wait a bit.
    for(i=0 ; i<10000000 ; i++);

    // Initializing before boot
    asm volatile ("mov r0,     #0");
    asm volatile ("mov r1,     #817&0xff");
    asm volatile ("orr r1, r1, #817&0xf00");
    asm volatile ("mov r2,     #0x10000000");
    asm volatile ("add r2, r2, #0x00000100");

    // Jump to the kernel.
    asm volatile ("mov r3,     #0x00008000");
    asm volatile ("orr r3, r3, #0x10000000");
    asm volatile ("mov pc, r3");

    // We should never get here... Just to make the compiler happy.
    return 0;
}
