/*
 * garux.h
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2005 Romain Goyet
 * Author: Romain Goyet [r.goyet@gmail.com]
 *
 * Original version by Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */


/* CAUTION : DO _never_ use the TOKEN below anywhere else in this file
 *  as this TOKEN is going to be dynamically replaced at compile time
 */


#ifndef __LOADLIN_H__
#define __LOADLIN_H__

//UI Definitions :

// Forms
#define MainForm 1000
#define AltForm 1001

//Menu
#define MainMenu 1000

//Menu items
#define MenuAbout 1000
#define MenuQuit 1001

//Alerts
#define StartupAlert 1000

//Buttons
#define LinuxButton 9998
#define PalmOSButton 9999

//Strings
#define GnuGpl 1000
#define kernelDefaultParam 1001

//Fields
#define KernelParam 1000

//Bitmaps
#define Splash	1005

#ifdef __GNUC__
typedef unsigned long uint32_t;
typedef unsigned short uint16_t;
typedef unsigned char uint8_t;

struct kernel_page_s {
	uint32_t magic;
	char count[4];
	uint8_t payload[0];
};

# define DEST_ADDR		0x11800000

struct fb_params_s {
	uint8_t *fb_address;
	uint32_t video_num_cols;
	uint32_t video_num_rows;
	uint32_t video_x;
	uint32_t video_y;
	uint32_t bytes_per_char_h;
	uint32_t bytes_per_char_v;
};

# define omap_writeb(x, y)	*(uint8_t *) (y) = x
# define omap_writew(x, y)	*(uint16_t *) (y) = x
# define omap_writel(x, y)	*(uint32_t *) (y) = x
# define omap_readb(y)		(*(uint8_t *) (y))
# define omap_readw(y)		(*(uint16_t *) (y))
# define omap_readl(y)		(*(uint32_t *) (y))
# define omap_setw(addr, bit, wdth) omap_writew(	\
		omap_readw(addr) | (((1 << (wdth)) - 1) << (bit)), addr)
# define omap_setl(addr, bit, wdth) omap_writel(	\
		omap_readl(addr) | (((1 << (wdth)) - 1) << (bit)), addr)
# define omap_clrw(addr, bit, wdth) omap_writew(	\
		omap_readw(addr) & ~(((1 << (wdth)) - 1) << (bit)), addr)
# define omap_clrl(addr, bit, wdth) omap_writel(	\
		omap_readl(addr) & ~(((1 << (wdth)) - 1) << (bit)), addr)

#endif

/* NUM_KERNEL_PAGES is the number of times PAGE_SIZE the kernel
 * will need. This is dynamically calculated at compile time
 */
#define NUM_KERNEL_PAGES TOKEN_TO_REPLACE
#define PAGE_SIZE 0x8000

#endif
