/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <MemGlue.h>
#include <MemoryMgr.h>
#include <string.h>
#include <stdio.h>

#include "garux.h"

  FieldType *kernelParametersField;

int run_palmte (void) {
  int i = 0;
  MemPtr kernel;
  void *kernelPointer;
  MemHandle asmloaderHandle;
  void *asmloader;
  MemHandle pageHandle;
  void *page;

	    // User choosed to load Linux

	    // Reserve a chunk of memory for the kernel
	    // We allocate 1Mb.
	    kernel = MemGluePtrNew(0x00100000);

	    // Fetch kernel command line
	    strcpy(kernel, FldGetTextPtr(kernelParametersField));
	    FldSetTextHandle(kernelParametersField, 0);

	    // Let's close the form now that we don't need the text field anymore
	    FrmCloseAllForms();
	    
	    // Let's resume memory allocation

	    kernelPointer = (void*)kernel+256;
	    asmloaderHandle = DmGetResource('armc', 0);
	    asmloader = MemHandleLock(asmloaderHandle);

	    printf("Loading kernel... at 0x%lx\n",(unsigned long)kernel);

	    
	    // Copy the kernel pages into the contiguous memory we allocated.
	    printf("Joining kernel chunks");
	    for (i = 0; i < NUM_KERNEL_PAGES; i++) {
	      pageHandle = DmGetResource('page', i);
	      page = MemHandleLock(pageHandle);
	      memcpy(kernelPointer, page, PAGE_SIZE);
	      kernelPointer += PAGE_SIZE;
	      MemHandleUnlock(pageHandle);
	      DmReleaseResource(pageHandle);
	      putchar('.');
	    }
	    printf("OK\n");
	    
	    printf("Running assembly loader...\n");
	    PceNativeCall(asmloader, kernel);
	    
	    // The following code shall not be executed
	    MemPtrFree(kernel);
	    MemHandleUnlock (asmloaderHandle);
	    return 0;
}

int run_palmtt2 (void) {
  int i = 0;
  Err retcode;
  UInt32 creator=3007;
  UInt16 featureNum=35;
  UInt32 offset=0;
  MemPtr kernel;
  void *kernelPointer;
  MemHandle asmloaderHandle;
  void *asmloader;
  MemHandle pageHandle;
  void *page;


	    //MemGluePtrNew() on palm tt2 allocate <750Kb,FtrPtrNew() allocate more

	    retcode = FtrPtrNew (creator,featureNum,0x200000,(void**)&kernel);
	    
	    if (retcode != 0) {
	    	printf ("Ftr not allocate memory\n");
		 SysTaskDelay( 1*SysTicksPerSecond() );
		 return -1;
	    }

	    // Fetch kernel command line
	    DmWrite(kernel, 0, FldGetTextPtr(kernelParametersField), 256);
	    FldSetTextHandle(kernelParametersField, 0);

	    // Let's close the form now that we don't need the text field anymore
	    FrmCloseAllForms();

	    kernelPointer = kernel;
	    asmloaderHandle = DmGetResource('armc', 1);
	    asmloader = MemHandleLock(asmloaderHandle);
	    //SysTaskDelay( 1*SysTicksPerSecond() );
	    
	    printf("Loading kernel... at 0x%lx\n",(unsigned long)kernel);

	    
	    // Copy the kernel pages into the contiguous memory we allocated.
	    printf("Joining kernel chunks");
	    for (i = 0; i < NUM_KERNEL_PAGES; i++) {
	      pageHandle = DmGetResource('page', i);
	      page = MemHandleLock(pageHandle);

	      DmWrite(kernelPointer, 256 + offset, page, PAGE_SIZE);
	      offset += PAGE_SIZE;

	      MemHandleUnlock(pageHandle);
	      DmReleaseResource(pageHandle);
	      putchar('.');
	    }
	    printf("OK\n");
	    
	    printf("Running assembly loader...\n");
	    PceNativeCall(asmloader, kernel);
	    
	    // The following code shall not be executed
	    FtrPtrFree(creator,featureNum);
	    MemHandleUnlock (asmloaderHandle);
	    return 0;
}


UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
  //  MemHandle defaultBootParm;
//  char *kernelParameters = "root=/dev/mmcblk0p1 mem=32m console=tty0 fbcon=map:0 font:MINI4x6";
  EventType event;
  UInt16 error_code;
  char deviceID[4];
  enum Palm_who { palm_undef, palm_te, palm_tt2 } who;

   
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    // Let's get the device ID in order to check wether we're on a T|E
    FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, (UInt32 *)&deviceID);

    if (deviceID[0] == 'C' && deviceID[1] == 'c' && deviceID[2] == 't' && deviceID[3] == '1')
	who=palm_te;
    else if (deviceID[0] == 'F' && deviceID[1] == 'r' && deviceID[2] == 'g' && deviceID[3] == '2')
	who=palm_tt2;
    else 
	who=palm_undef;

    if (who!=palm_undef) {
      // Garux is being run on a Tungsten E

      FrmGotoForm(MainForm);
      while (1) {
	EvtGetEvent(&event, 100);
	if (SysHandleEvent(&event))
	  continue;
	if (MenuHandleEvent(0, &event, &error_code))
	  continue;
	if (event.eType == ctlSelectEvent) {
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FldSetTextHandle(kernelParametersField, 0);
	    FrmCloseAllForms();
	    return 0;
	  }
	  else if (event.data.ctlSelect.controlID == LinuxButton) {
	  
	  switch (who) {
	  case palm_te:
	    run_palmte();
	    break;
	  case palm_tt2:
	    run_palmtt2();
	    break;
	  default:
	    break;
	  }

	    return 0;
	  }
	}
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());

	  // Initial warning
	  if (FrmAlert(StartupAlert) == 1) {
	    FrmCloseAllForms();
	    return 0;
	  }

	  // Initializing boot parameters
	  kernelParametersField = (FieldType *)FrmGetObjectPtr(FrmGetFormPtr(MainForm), FrmGetObjectIndex(FrmGetFormPtr(MainForm), KernelParam));

	  FldSetTextHandle(kernelParametersField, DmGetResource(strRsc, kernelDefaultParam));
	  FldRecalculateField(kernelParametersField, true);

	  continue;
	}
	if (event.eType == appStopEvent) {
	  FldSetTextHandle(kernelParametersField, 0);
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	  FrmHandleEvent(FrmGetActiveForm(), &event);
      }
    }
    else {
      // Garux is not being run on a Tungsten E. We'll warn, and exit.
      FrmGotoForm(AltForm);
      
      while (1) {
	EvtGetEvent(&event, 100);
	SysHandleEvent(&event);
	if (event.eType == ctlSelectEvent)
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FrmCloseAllForms();
	    return 0;
	  }
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());
	  continue;
	}
	if (event.eType == appStopEvent) {
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	FrmHandleEvent(FrmGetActiveForm(), &event); 
      }
    }
    return 0;
  }
}
