/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <MemGlue.h>
#include <MemoryMgr.h>
#include <string.h>
#include <stdio.h>

#include "garux.h"


UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
  EventType event;
  UInt16 error_code;
  char deviceID[4];
  MemPtr cmdline;
  MemHandle asmloaderHandle;
  void *asmloader;
  FieldType *kernelParametersField = 0;
  MemHandle splashRes;
  BitmapType *splashBmp;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    // Let's get the device ID in order to check wether we're on a T|E
    FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, (UInt32 *)&deviceID);

    if (deviceID[0] == 'C' && deviceID[1] == 'c' && deviceID[2] == 't' && deviceID[3] == '1') {
      // Garux is being run on a Tungsten E

      FrmGotoForm(MainForm);
      while (1) {
	EvtGetEvent(&event, 100);
	if (SysHandleEvent(&event))
	  continue;
	if (MenuHandleEvent(0, &event, &error_code))
	  continue;
	if (event.eType == ctlSelectEvent) {
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FldSetTextHandle(kernelParametersField, 0);
	    FrmCloseAllForms();
	    return 0;
	  }
	  else if (event.data.ctlSelect.controlID == LinuxButton) {
	    // User chose to load Linux

	    // Reserve a chunk of memory for the commandline
	    cmdline = MemGluePtrNew(
			    strlen(FldGetTextPtr(kernelParametersField)) + 1);

	    // Fetch kernel command line
	    strcpy(cmdline, FldGetTextPtr(kernelParametersField));
	    FldSetTextHandle(kernelParametersField, 0);

	    // Let's close the form now that we don't need the text field anymore
	    FrmCloseAllForms();

	    printf("Commandline loaded at 0x%lx.\n", (unsigned long) cmdline);

	    printf("Joining kernel chunks deferred.\n");

	    asmloaderHandle = DmGetResource('armc', 0);
	    asmloader = MemHandleLock(asmloaderHandle);

	    printf("Running assembly loader...\n");

#ifdef SPLASH
	    splashRes = DmGetResource(bitmapRsc, Splash);
	    splashBmp = MemHandleLock(splashRes);
	    WinPaintBitmap(splashBmp, 0, 0);
	    MemHandleUnlock(splashRes);
#endif

	    PceNativeCall(asmloader, cmdline);

	    // The following code shall not be executed
	    MemPtrFree(cmdline);
	    MemHandleUnlock(asmloaderHandle);
	    return 0;
	  }
	}
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());

	  // Initial warning
	  if (FrmAlert(StartupAlert) == 1) {
	    FrmCloseAllForms();
	    return 0;
	  }

	  // Initializing boot parameters
	  kernelParametersField = (FieldType *)FrmGetObjectPtr(FrmGetFormPtr(MainForm), FrmGetObjectIndex(FrmGetFormPtr(MainForm), KernelParam));

	  FldSetTextHandle(kernelParametersField, DmGetResource(strRsc, kernelDefaultParam));
	  FldRecalculateField(kernelParametersField, true);

	  continue;
	}
	if (event.eType == appStopEvent) {
	  FldSetTextHandle(kernelParametersField, 0);
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	  FrmHandleEvent(FrmGetActiveForm(), &event);
      }
    }
    else {
      // Garux is not being run on a Tungsten E. We'll warn, and exit.
      FrmGotoForm(AltForm);
      
      while (1) {
	EvtGetEvent(&event, 100);
	SysHandleEvent(&event);
	if (event.eType == ctlSelectEvent)
	  if (event.data.ctlSelect.controlID == PalmOSButton) {
	    FrmCloseAllForms();
	    return 0;
	  }
	if (event.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
	  continue;
	}
	if (event.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());
	  continue;
	}
	if (event.eType == appStopEvent) {
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	FrmHandleEvent(FrmGetActiveForm(), &event); 
      }
    }
  }
  return 0;
}
