/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <MemGlue.h>
#include <MemoryMgr.h>
#include <string.h>
#include <stdio.h>

#include "garux.h"

#define u32			UInt32

#define OMAP_DIE_ID_0		0xfffe1800
#define OMAP_DIE_ID_1		0xfffe1804
#define OMAP_PRODUCTION_ID_0	0xfffe2000
#define OMAP_PRODUCTION_ID_1	0xfffe2004
#define OMAP32_ID_0		0xfffed400
#define OMAP32_ID_1		0xfffed404

#define IO_PHYS			0xFFFB0000
#define IO_OFFSET		0x01000000	/* Virtual IO = 0xfefb0000 */
#define IO_VIRT			(IO_PHYS - IO_OFFSET)
#define IO_SIZE			0x40000
#define IO_ADDRESS(x)		((x) - IO_OFFSET)
#define omap_readl(a)		(*(volatile unsigned int   *)IO_ADDRESS(a))

UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
  int i = 0;
  EventType e;
  char deviceID[4];
  MemPtr kernel;
  void *kernelPointer;
  MemHandle asmloaderHandle;
  void *asmloader;
  MemHandle pageHandle;
  void *page;
	u32 prod_id0, prod_id1;
	u32 omap_id0, omap_id1;
	u32 die_rev0, die_rev1;

	prod_id0 = omap_readl(OMAP_PRODUCTION_ID_0);
	prod_id1 = omap_readl(OMAP_PRODUCTION_ID_1);
	omap_id0 = omap_readl(OMAP32_ID_0);
	omap_id1 = omap_readl(OMAP32_ID_1);
	die_rev0 = omap_readl(OMAP_DIE_ID_0);
	die_rev1 = omap_readl(OMAP_DIE_ID_1);

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    // Let's get the device ID in order to check wether we're on a T|E
    FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, (UInt32 *)&deviceID);

    if (deviceID[0] == 'C' && deviceID[1] == 'c' && deviceID[2] == 't' && deviceID[3] == '1') {
      // Garux is being run on a Tungsten E
      FrmGotoForm(MainForm);
      while (1) {
	EvtGetEvent(&e, 100);
	SysHandleEvent(&e);
	if (e.eType == ctlSelectEvent) {
	  if (e.data.ctlSelect.controlID == PalmOSButton) {
	    FrmCloseAllForms();
	    return 0;
	  }
	  else if (e.data.ctlSelect.controlID == LinuxButton) {
	    // User choosed to load Linux
	    
	    // Let's clean the screen first
	    FrmCloseAllForms();
	    
	    // Reserve a chumk of memory for the kernel
	    // We allocate 1Mb.
	    
	    printf("Production ID:  %lx %lx\n", prod_id0, prod_id1);
	    printf("Omap ID:        %lx %lx\n", omap_id0, omap_id1);
	    printf("Revision ID:    %lx %lx\n", die_rev0, die_rev1);

/*	    printf("Running assembly loader...\n");
	    PceNativeCall(asmloader, kernel);
	    
	    // The following code shall not be executed
	    MemPtrFree(kernel);
	    MemHandleUnlock (asmloaderHandle);*/
	   /* return 0;*/
	  }
	}
	if (e.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(e.data.frmLoad.formID));
	  continue;
	}
	if (e.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());
	  continue;
	}
	if (e.eType == appStopEvent) {
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	  FrmHandleEvent(FrmGetActiveForm(), &e);
      }
    }
    else {
      // Garux is not being run on a Tungsten E. We'll warn, and exit.
      FrmGotoForm(AltForm);
      
      while (1) {
	EvtGetEvent(&e, 100);
	SysHandleEvent(&e);
	if (e.eType == ctlSelectEvent)
	  if (e.data.ctlSelect.controlID == PalmOSButton) {
	    FrmCloseAllForms();
	    return 0;
	  }
	if (e.eType == frmLoadEvent) {
	  FrmSetActiveForm(FrmInitForm(e.data.frmLoad.formID));
	  continue;
	}
	if (e.eType == frmOpenEvent) {
	  FrmDrawForm(FrmGetActiveForm());
	  continue;
	}
	if (e.eType == appStopEvent) {
	  FrmCloseAllForms();
	  return 0;	
	}
	if (FrmGetActiveForm())
	FrmHandleEvent(FrmGetActiveForm(), &e); 
      }
    }
    return 0;
  }
}
